import json
import logging
import os
import warnings

from google.oauth2 import service_account


logger = logging.getLogger(__name__)


class GCPCredentialsManager:
    """A credentials manager for Google Cloud Platform (GCP) that takes a path to a service account JSON file, or a
    JSON string of the contents of such a service account file, from the given environment variable and instantiates
    a Google Cloud credentials object.

    :param str|None environment_variable_name:
    :return None:
    """

    def __init__(self, environment_variable_name="GOOGLE_APPLICATION_CREDENTIALS"):
        self.environment_variable_name = environment_variable_name

        if self.environment_variable_name is None:
            self.service_account_json = None
            return

        try:
            self.service_account_json = os.environ[self.environment_variable_name]
        except KeyError:
            warnings.warn(
                f"No environment variable called {self.environment_variable_name!r}; resorting to default Google Cloud "
                f"credentials."
            )
            self.service_account_json = None

    def get_credentials(self, as_dict=False):
        """Get the Google OAUTH2 service account credentials.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials|None:
        """
        if self.service_account_json is None:
            return None

        # Check that the environment variable refers to a real file.
        if os.path.exists(self.service_account_json):
            return self._get_credentials_from_file(as_dict=as_dict)

        # If it doesn't, assume that it's the credentials file as a JSON string.
        return self._get_credentials_from_string(as_dict=as_dict)

    def _get_credentials_from_file(self, as_dict=False):
        """Get the credentials from the JSON file whose path is specified in the environment variable's value.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials:
        """
        with open(self.service_account_json) as f:
            credentials = json.load(f)

        logger.debug("GCP credentials read from file.")

        if as_dict:
            return credentials

        return service_account.Credentials.from_service_account_info(credentials)

    def _get_credentials_from_string(self, as_dict=False):
        """Get the credentials directly from the JSON string specified in the environment variable's value.

        :param bool as_dict: if `True`, get the credentials as a dictionary
        :return dict|google.auth.service_account.Credentials:
        """
        credentials = json.loads(self.service_account_json)
        logger.debug("GCP credentials loaded from string.")

        if as_dict:
            return credentials

        return service_account.Credentials.from_service_account_info(credentials)



import logging

from octue.resources.analysis import HASH_FUNCTIONS, Analysis
from twined import Twine

from ..base import BaseTestCase


class AnalysisTestCase(BaseTestCase):
    def test_instantiate_analysis(self):
        """Ensures that the base analysis class can be instantiated"""
        # from octue import runner  # <-- instantiated in the library, not here
        analysis = Analysis(twine="{}")
        self.assertEqual(analysis.__class__.__name__, "Analysis")

    def test_instantiate_analysis_with_twine(self):
        """Ensures that the base analysis class can be instantiated"""
        analysis = Analysis(twine=Twine(source="{}"))
        self.assertEqual(analysis.__class__.__name__, "Analysis")

    def test_non_existent_attributes_cannot_be_retrieved(self):
        """Ensure attributes that don't exist on Analysis aren't retrieved as None and instead raise an error. See
        https://github.com/octue/octue-sdk-python/issues/45 for reasoning behind adding this.
        """
        analysis = Analysis(twine=Twine(source="{}"))

        with self.assertRaises(AttributeError):
            analysis.furry_purry_cat

    def test_analysis_hash_attributes_are_none_when_no_relevant_strands(self):
        """Ensures that the hash attributes of Analysis instances are None if none of the relevant strands are provided"""
        analysis = Analysis(twine="{}")
        for strand_name in HASH_FUNCTIONS:
            self.assertIsNone(getattr(analysis, f"{strand_name}_hash"))

    def test_analysis_hash_attributes_are_populated_when_relevant_strands_are_present(self):
        """Ensures that the hash attributes of Analysis instances are valid if the relevant strands are provided."""
        analysis = Analysis(
            twine="{}",
            configuration_values={"resistance_setting": 7},
            configuration_manifest=self.create_valid_manifest(),
            input_values={"quality_factor": 5},
            input_manifest=self.create_valid_manifest(),
        )

        for strand_name in HASH_FUNCTIONS:
            hash_ = getattr(analysis, f"{strand_name}_hash")
            self.assertTrue(isinstance(hash_, str))
            self.assertTrue(len(hash_) == 8)

    def test_warning_raised_if_attempting_to_send_a_monitoring_update_but_no_monitoring_callback_is_provided(self):
        """Test that a warning is raised if attempting to send a monitoring update but no monitoring callback is
        provided.
        """
        analysis = Analysis(twine='{"monitor_message_schema": {}}')

        with self.assertLogs(level=logging.WARNING) as logging_context:
            analysis.send_monitor_message(data=[])

        self.assertIn(
            "Attempted to send a monitor message but no handler is specified.",
            logging_context.output[0],
        )



import json
import os
import tempfile
from unittest.mock import patch

import google.oauth2.service_account

from octue.cloud.credentials import GCPCredentialsManager
from tests.base import BaseTestCase


class TestGCPCredentialsManager(BaseTestCase):
    def test_warning_issued_if_no_environment_variable(self):
        """Ensure a warning is issued if the given environment variable can't be found."""
        with self.assertWarns(Warning):
            GCPCredentialsManager(environment_variable_name="heeby_jeeby-11111-33103")

    def test_credentials_can_be_loaded_from_file(self):
        """Test that credentials can be loaded from the file at the path specified by the environment variable."""
        credentials = GCPCredentialsManager().get_credentials()
        self.assertEqual(type(credentials), google.oauth2.service_account.Credentials)

    def test_credentials_can_be_loaded_as_dictionary_from_string(self):
        """Test that credentials can be loaded as a dictionary from a string-valued environment variable."""
        with patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": '{"blah": "nah"}'}):
            credentials = GCPCredentialsManager().get_credentials(as_dict=True)
            self.assertEqual(credentials, {"blah": "nah"})

    def test_credentials_can_be_loaded_as_dictionary_from_file(self):
        """Test that credentials can be loaded as a dictionary from a file."""
        with tempfile.NamedTemporaryFile(delete=False) as temporary_file:

            with open(temporary_file.name, "w") as f:
                json.dump({"blah": "nah"}, f)

            with patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": temporary_file.name}):
                credentials = GCPCredentialsManager().get_credentials(as_dict=True)
                self.assertEqual(credentials, {"blah": "nah"})

    def test_credentials_are_none_when_environment_variable_name_is_none(self):
        """Test that credentials are `None` when the given environment variable name is `None`."""
        self.assertIsNone(GCPCredentialsManager(environment_variable_name=None).get_credentials())



"""A module containing helper functions for sending monitor messages that conform to the Octue essential monitor
message schema https://refs.schema.octue.com/octue/essential-monitors/0.0.2.json
"""
import logging
from datetime import datetime, timezone


logger = logging.getLogger(__name__)


def send_status_text(analysis, text, service_name):
    """Send a status-type monitor message and additionally log it to the info level.

    :param octue.resources.analysis.Analysis analysis: the analysis from which to send the status text
    :param str text: the text of the status message
    :param str service_name: the name of the service/child running the analysis
    :return None:
    """
    analysis.send_monitor_message(
        {"service": service_name, "status_text": text, "date_time": datetime.now(timezone.utc).isoformat()}
    )
    logger.info(text)


def send_estimated_seconds_remaining(analysis, estimated_seconds_remaining, service_name):
    """Send an estimated-seconds-remaining monitor message.

    :param octue.resources.analysis.Analysis analysis: the analysis from which to send the estimate
    :param float estimated_seconds_remaining:
    :param str service_name: the name of the service/child running the analysis
    """
    analysis.send_monitor_message(
        {
            "service": service_name,
            "estimated_seconds_remaining": estimated_seconds_remaining,
            "date_time": datetime.now(timezone.utc).isoformat(),
        }
    )



import json
import os
import warnings

import octue.migrations.manifest
from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import InvalidInputException
from octue.mixins import Hashable, Identifiable, Pathable, Serialisable

from .dataset import Dataset


class Manifest(Pathable, Serialisable, Identifiable, Hashable):
    """A representation of a manifest, which can contain multiple datasets This is used to manage all files coming into
    (or leaving), a data service for an analysis at the configuration, input or output stage.

    :param str|None id: the UUID of the manifest (a UUID is generated if one isn't given)
    :param str|None path: the path the manifest exists at (defaults to the current working directory)
    :param dict(str, octue.resources.dataset.Dataset|dict|str)|None datasets: a mapping of dataset names to `Dataset` instances, serialised datasets, or paths to datasets
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("datasets",)
    _SERIALISE_FIELDS = "datasets", "keys", "id", "name", "path"

    def __init__(self, id=None, path=None, datasets=None, **kwargs):
        if isinstance(datasets, list):
            datasets = octue.migrations.manifest.translate_datasets_list_to_dictionary(datasets, kwargs.get("keys"))

        super().__init__(id=id, path=path)
        self.datasets = {}

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then tidy up this hackjob. Also need to allow Pathables to update ownership
        #  (because decoders work from the bottom of the tree upwards, not top-down)

        self._instantiate_datasets(datasets or {})
        vars(self).update(**kwargs)

    @classmethod
    def from_cloud(cls, project_name=None, cloud_path=None, bucket_name=None, path_to_manifest_file=None):
        """Instantiate a Manifest from Google Cloud storage. Either (`bucket_name` and `path_to_manifest_file`) or
        `cloud_path` must be provided.

        :param str|None cloud_path: full path to manifest in cloud storage (e.g. `gs://bucket_name/path/to/manifest.json`)
        :param str|None bucket_name: name of bucket manifest is stored in
        :param str|None path_to_manifest_file: path to manifest in cloud storage e.g. `path/to/manifest.json`
        :return Dataset:
        """
        if project_name:
            warnings.warn(
                message="The `project_name` parameter is no longer needed and will be removed soon.",
                category=DeprecationWarning,
            )

        if cloud_path:
            bucket_name, path_to_manifest_file = storage.path.split_bucket_name_from_gs_path(cloud_path)

        serialised_manifest = json.loads(
            GoogleCloudStorageClient().download_as_string(bucket_name=bucket_name, path_in_bucket=path_to_manifest_file)
        )

        datasets = {}

        for key, dataset in serialised_manifest["datasets"].items():
            dataset_bucket_name, path = storage.path.split_bucket_name_from_gs_path(dataset)
            datasets[key] = Dataset.from_cloud(bucket_name=dataset_bucket_name, path_to_dataset_directory=path)

        return Manifest(
            id=serialised_manifest["id"],
            path=storage.path.generate_gs_path(bucket_name, path_to_manifest_file),
            datasets=datasets,
        )

    def to_cloud(
        self, project_name=None, cloud_path=None, bucket_name=None, path_to_manifest_file=None, store_datasets=True
    ):
        """Upload a manifest to a cloud location, optionally uploading its datasets into the same directory. Either
        (`bucket_name` and `path_to_manifest_file`) or `cloud_path` must be provided.

        :param str|None cloud_path: full path to cloud storage location to store manifest at (e.g. `gs://bucket_name/path/to/manifest.json`)
        :param str|None bucket_name: name of bucket to store manifest in
        :param str|None path_to_manifest_file: cloud storage path to store manifest at e.g. `path/to/manifest.json`
        :param bool store_datasets: if True, upload datasets to same directory as manifest file
        :return str: gs:// path for manifest file
        """
        if project_name:
            warnings.warn(
                message="The `project_name` parameter is no longer needed and will be removed soon.",
                category=DeprecationWarning,
            )

        if cloud_path:
            bucket_name, path_to_manifest_file = storage.path.split_bucket_name_from_gs_path(cloud_path)

        datasets = {}
        output_directory = storage.path.dirname(path_to_manifest_file)

        for key, dataset in self.datasets.items():
            if store_datasets:
                dataset_path = dataset.to_cloud(bucket_name=bucket_name, output_directory=output_directory)
                datasets[key] = dataset_path
            else:
                datasets[key] = dataset.absolute_path

        serialised_manifest = self.to_primitive()
        serialised_manifest["datasets"] = datasets
        del serialised_manifest["path"]

        GoogleCloudStorageClient().upload_from_string(
            string=json.dumps(serialised_manifest),
            bucket_name=bucket_name,
            path_in_bucket=path_to_manifest_file,
        )

        return cloud_path or storage.path.generate_gs_path(bucket_name, path_to_manifest_file)

    @property
    def all_datasets_are_in_cloud(self):
        """Do all the files of all the datasets of the manifest exist in the cloud?

        :return bool:
        """
        if not self.datasets:
            return False

        return all(dataset.all_files_are_in_cloud for dataset in self.datasets.values())

    def get_dataset(self, key):
        """Get a dataset by its key (as defined in the twine).

        :param str key:
        :return octue.resources.dataset.Dataset:
        """
        dataset = self.datasets.get(key, None)

        if dataset is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset {key!r} from Manifest. Allowable keys are: "
                f"{list(self.datasets.keys())}"
            )

        return dataset

    def prepare(self, data):
        """Prepare new manifest from a manifest_spec.

        :param dict data:
        :return Manifest:
        """
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for key, dataset_specification in data["datasets"].items():
            # TODO generate a unique name based on the filter key, label datasets so that the label filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets[key] = Dataset(path_from=self, path=key)

        return self

    def _instantiate_datasets(self, datasets):
        """Add the given datasets to the manifest, instantiating them if needed and giving them the correct path.
        There are several possible forms the datasets can come in:
        * Instantiated Dataset instances
        * A list of dictionaries pointing to cloud datasets
        * Fully serialised form - includes path
        * `manifest.json` form - does not include path
        * Including datafiles that already exist
        * Including datafiles that don't yet exist or are not possessed currently (e.g. future output locations or
          cloud files)

        :param dict(str, octue.resources.dataset.Dataset|dict|str) datasets: the datasets to add to the manifest
        :return None:
        """
        for key, dataset in datasets.items():

            if isinstance(dataset, Dataset):
                self.datasets[key] = dataset

            else:
                # If `dataset` is just a path to a dataset:
                if isinstance(dataset, str):
                    if storage.path.is_qualified_cloud_path(dataset):
                        self.datasets[key] = Dataset.from_cloud(cloud_path=dataset, recursive=True)
                    else:
                        self.datasets[key] = Dataset.from_local_directory(path_to_directory=dataset, recursive=True)

                # If `dataset` is a dictionary including a "path" key:
                elif "path" in dataset:
                    # If the path is not a cloud path or an absolute local path:
                    if not os.path.isabs(dataset["path"]) and not storage.path.is_qualified_cloud_path(dataset["path"]):
                        path = dataset.pop("path")
                        self.datasets[key] = Dataset(**dataset, path=path, path_from=self)

                    # If the path is a cloud path or an absolute local path:
                    else:
                        self.datasets[key] = Dataset(**dataset)

                else:
                    self.datasets[key] = Dataset(**dataset, path=key, path_from=self)



import base64
import concurrent.futures
import datetime
import functools
import json
import logging
import time
import uuid

from google.api_core import retry
from google.cloud import pubsub_v1

import octue.exceptions
from octue.cloud.credentials import GCPCredentialsManager
from octue.cloud.pub_sub import Subscription, Topic
from octue.cloud.pub_sub.logging import GooglePubSubHandler
from octue.cloud.pub_sub.message_handler import OrderedMessageHandler
from octue.mixins import CoolNameable
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.exceptions import convert_exception_to_primitives
from octue.utils.objects import get_nested_attribute


logger = logging.getLogger(__name__)

OCTUE_NAMESPACE = "octue.services"
ANSWERS_NAMESPACE = "answers"

# Switch message batching off by setting max_messages to 1. This minimises latency and is recommended for
# microservices publishing single messages in a request-response sequence.
BATCH_SETTINGS = pubsub_v1.types.BatchSettings(max_bytes=10 * 1000 * 1000, max_latency=0.01, max_messages=1)


class Service(CoolNameable):
    """A Twined service that can be used in two modes:
    * As a server accepting questions (input values and manifests), running them through its app, and responding to the
    requesting service with the results of the analysis.
    * As a requester of answers from another Service in the above mode.

    Services communicate entirely via Google Pub/Sub and can ask and/or respond to questions from any other Service that
    has a corresponding topic on Google Pub/Sub.

    :param octue.resources.service_backends.ServiceBackend backend: the object representing the type of backend the service uses
    :param str|None service_id: a string UUID optionally preceded by the octue services namespace "octue.services."
    :param callable|None run_function: the function the service should run when it is called
    :return None:
    """

    def __init__(self, backend, service_id=None, run_function=None, *args, **kwargs):
        if service_id is None:
            self.id = f"{OCTUE_NAMESPACE}.{str(uuid.uuid4())}"
        elif not service_id:
            raise ValueError(f"service_id should be None or a non-falsey value; received {service_id!r} instead.")
        else:
            if service_id.startswith(OCTUE_NAMESPACE):
                self.id = service_id
            else:
                self.id = f"{OCTUE_NAMESPACE}.{service_id}"

        self.backend = backend
        self.run_function = run_function

        self._credentials = GCPCredentialsManager(backend.credentials_environment_variable).get_credentials()
        self.publisher = pubsub_v1.PublisherClient(credentials=self._credentials, batch_settings=BATCH_SETTINGS)
        self._current_question = None
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"<{type(self).__name__}({self.name!r})>"

    def serve(self, timeout=None, delete_topic_and_subscription_on_exit=False):
        """Start the Service as a server, waiting to accept questions from any other Service using Google Pub/Sub on
        the same Google Cloud Platform project. Questions are responded to asynchronously.

        :param float|None timeout: time in seconds after which to shut down the service
        :param bool delete_topic_and_subscription_on_exit: if `True`, delete the service's topic and subscription on exit
        :return None:
        """
        topic = Topic(name=self.id, namespace=OCTUE_NAMESPACE, service=self)
        topic.create(allow_existing=True)

        subscriber = pubsub_v1.SubscriberClient(credentials=self._credentials)

        subscription = Subscription(
            name=self.id,
            topic=topic,
            namespace=OCTUE_NAMESPACE,
            project_name=self.backend.project_name,
            subscriber=subscriber,
            expiration_time=None,
        )
        subscription.create(allow_existing=True)

        future = subscriber.subscribe(subscription=subscription.path, callback=self.answer)
        logger.debug("%r is waiting for questions.", self)

        with subscriber:
            try:
                future.result(timeout=timeout)
            except (TimeoutError, concurrent.futures.TimeoutError, KeyboardInterrupt):
                future.cancel()

            if delete_topic_and_subscription_on_exit:
                topic.delete()
                subscription.delete()

    def answer(self, question, answer_topic=None, timeout=30):
        """Answer a question (i.e. run the Service's app to analyse the given data, and return the output values to the
        asker). Answers are published to a topic whose name is generated from the UUID sent with the question, and are
        in the format specified in the Service's Twine file.

        :param dict|Message question:
        :param octue.cloud.pub_sub.topic.Topic|None answer_topic: provide if messages need to be sent to the asker from outside the service (e.g. in octue.cloud.deployment.google.cloud_run.flask_app)
        :param float|None timeout: time in seconds to keep retrying sending of the answer once it has been calculated
        :raise Exception: if any exception arises during running analysis and sending its results
        :return None:
        """
        data, question_uuid, forward_logs = self._parse_question(question)
        topic = answer_topic or self.instantiate_answer_topic(question_uuid)
        self._send_delivery_acknowledgment(topic)

        if forward_logs:
            analysis_log_handler = GooglePubSubHandler(publisher=self.publisher, topic=topic, analysis_id=question_uuid)
        else:
            analysis_log_handler = None

        try:
            analysis = self.run_function(
                analysis_id=question_uuid,
                input_values=data["input_values"],
                input_manifest=data["input_manifest"],
                analysis_log_handler=analysis_log_handler,
                handle_monitor_message=functools.partial(self._send_monitor_message, topic=topic),
            )

            if analysis.output_manifest is None:
                serialised_output_manifest = None
            else:
                serialised_output_manifest = analysis.output_manifest.serialise()

            self.publisher.publish(
                topic=topic.path,
                data=json.dumps(
                    {
                        "type": "result",
                        "output_values": analysis.output_values,
                        "output_manifest": serialised_output_manifest,
                        "message_number": topic.messages_published,
                    },
                    cls=OctueJSONEncoder,
                ).encode(),
                retry=retry.Retry(deadline=timeout),
            )
            topic.messages_published += 1
            logger.info("%r responded to question %r.", self, question_uuid)

        except BaseException as error:  # noqa
            self.send_exception_to_asker(topic, timeout)
            raise error

    def instantiate_answer_topic(self, question_uuid, service_id=None):
        """Instantiate the answer topic for the given question UUID for the given service ID.

        :param str question_uuid:
        :param str|None service_id: the ID of the service to ask the question to
        :return octue.cloud.pub_sub.topic.Topic:
        """
        return Topic(
            name=".".join((service_id or self.id, ANSWERS_NAMESPACE, question_uuid)),
            namespace=OCTUE_NAMESPACE,
            service=self,
        )

    def ask(
        self,
        service_id,
        input_values=None,
        input_manifest=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        timeout=30,
        question_uuid=None,
    ):
        """Ask a serving Service a question (i.e. send it input values for it to run its app on). The input values must
        be in the format specified by the serving Service's Twine file. A single-use topic and subscription are created
        before sending the question to the serving Service - the topic is the expected publishing place for the answer
        from the serving Service when it comes, and the subscription is set up to subscribe to this.

        :param str service_id: the UUID of the service to ask the question to
        :param any input_values: the input values of the question
        :param octue.resources.manifest.Manifest|None input_manifest: the input manifest of the question
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the remote service and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the serving service will have access to these local files
        :param float|None timeout: time in seconds to keep retrying sending the question
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :return (octue.cloud.pub_sub.subscription.Subscription, str): the response subscription and question UUID
        """
        if not allow_local_files:
            if (input_manifest is not None) and (not input_manifest.all_datasets_are_in_cloud):
                raise octue.exceptions.FileLocationError(
                    "All datasets of the input manifest and all files of the datasets must be uploaded to the cloud "
                    "before asking a service to perform an analysis upon them. The manifest must then be updated with "
                    "the new cloud locations."
                )

        question_topic = Topic(name=service_id, namespace=OCTUE_NAMESPACE, service=self)

        if not question_topic.exists(timeout=timeout):
            raise octue.exceptions.ServiceNotFound(f"Service with ID {service_id!r} cannot be found.")

        # If a question UUID is given, this is probably a retry so allow the question topic to already exist.
        if question_uuid:
            allow_existing_topic = True
        else:
            allow_existing_topic = False

        question_uuid = question_uuid or str(uuid.uuid4())

        response_topic = self.instantiate_answer_topic(question_uuid, service_id)
        response_topic.create(allow_existing=allow_existing_topic)

        response_subscription = Subscription(
            name=response_topic.name,
            topic=response_topic,
            namespace=OCTUE_NAMESPACE,
            project_name=self.backend.project_name,
            subscriber=pubsub_v1.SubscriberClient(credentials=self._credentials),
        )
        response_subscription.create(allow_existing=True)

        serialised_input_manifest = None
        if input_manifest is not None:
            serialised_input_manifest = input_manifest.serialise()

        self.publisher.publish(
            topic=question_topic.path,
            data=json.dumps({"input_values": input_values, "input_manifest": serialised_input_manifest}).encode(),
            question_uuid=question_uuid,
            forward_logs=str(int(subscribe_to_logs)),
            retry=retry.Retry(deadline=timeout),
        )

        # Keep a record of the question asked in case it needs to be retried.
        self._current_question = {
            "service_id": service_id,
            "input_values": input_values,
            "input_manifest": input_manifest,
            "question_uuid": question_uuid,
            "subscribe_to_logs": subscribe_to_logs,
            "allow_local_files": allow_local_files,
            "timeout": timeout,
        }

        logger.info("%r asked a question %r to service %r.", self, question_uuid, service_id)
        return response_subscription, question_uuid

    def wait_for_answer(
        self,
        subscription,
        handle_monitor_message=None,
        service_name="REMOTE",
        timeout=60,
        delivery_acknowledgement_timeout=30,
        retry_interval=5,
    ):
        """Wait for an answer to a question on the given subscription, deleting the subscription and its topic once
        the answer is received.

        :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription for the question's answer
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param str service_name: an arbitrary name to refer to the service subscribed to by (used for labelling its remote log messages)
        :param float|None timeout: how long in seconds to wait for an answer before raising a `TimeoutError`
        :param float delivery_acknowledgement_timeout: how long in seconds to wait for a delivery acknowledgement before resending the question
        :param float retry_interval: the time in seconds to wait between question retries
        :raise TimeoutError: if the timeout is exceeded
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        subscriber = pubsub_v1.SubscriberClient(credentials=self._credentials)

        message_handler = OrderedMessageHandler(
            subscriber=subscriber,
            subscription=subscription,
            handle_monitor_message=handle_monitor_message,
            service_name=service_name,
        )

        with subscriber:

            try:
                # Retry sending the question until the overall timeout is reached.
                while not message_handler.received_delivery_acknowledgement:

                    try:
                        return message_handler.handle_messages(
                            timeout=timeout,
                            delivery_acknowledgement_timeout=delivery_acknowledgement_timeout,
                        )

                    except octue.exceptions.QuestionNotDelivered:
                        logger.info(
                            "%r: No acknowledgement of question delivery after %fs - resending in %fs.",
                            self,
                            delivery_acknowledgement_timeout,
                            retry_interval,
                        )

                        time.sleep(retry_interval)
                        self.ask(**self._current_question)

            finally:
                subscription.delete()

    def _send_delivery_acknowledgment(self, topic, timeout=30):
        """Send an acknowledgement of question delivery to the asker.

        :param octue.cloud.pub_sub.topic.Topic topic: topic to send acknowledgement to
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        logger.info("%r acknowledged receipt of question.", self)

        self.publisher.publish(
            topic=topic.path,
            data=json.dumps(
                {
                    "type": "delivery_acknowledgement",
                    "delivery_time": str(datetime.datetime.now()),
                    "message_number": topic.messages_published,
                }
            ).encode(),
            retry=retry.Retry(deadline=timeout),
        )

        topic.messages_published += 1

    def _send_monitor_message(self, data, topic, timeout=30):
        """Send a monitor message to the asker.

        :param any data: the data to send as a monitor message
        :param octue.cloud.pub_sub.topic.Topic topic: the topic to send the message to
        :param float timeout: time in seconds to retry sending the message
        :return None:
        """
        logger.debug("%r sending monitor message.", self)

        self.publisher.publish(
            topic=topic.path,
            data=json.dumps(
                {
                    "type": "monitor_message",
                    "data": json.dumps(data),
                    "message_number": topic.messages_published,
                }
            ).encode(),
            retry=retry.Retry(deadline=timeout),
        )

        topic.messages_published += 1

    def send_exception_to_asker(self, topic, timeout=30):
        """Serialise and send the exception being handled to the asker.

        :param octue.cloud.pub_sub.topic.Topic topic:
        :param float|None timeout: time in seconds to keep retrying sending of the exception
        :return None:
        """
        exception = convert_exception_to_primitives()
        exception_message = f"Error in {self!r}: {exception['message']}"

        self.publisher.publish(
            topic=topic.path,
            data=json.dumps(
                {
                    "type": "exception",
                    "exception_type": exception["type"],
                    "exception_message": exception_message,
                    "traceback": exception["traceback"],
                    "message_number": topic.messages_published,
                }
            ).encode(),
            retry=retry.Retry(deadline=timeout),
        )

        topic.messages_published += 1

    def _parse_question(self, question):
        """Parse a question in the Google Cloud Pub/Sub or Google Cloud Run format.

        :param dict|Message question:
        :return (dict, str, bool):
        """
        try:
            # Parse question directly from Pub/Sub or Dataflow.
            data = json.loads(question.data.decode())

            # Acknowledge it if it's directly from Pub/Sub
            if hasattr(question, "ack"):
                question.ack()

        except Exception:
            # Parse question from Google Cloud Run.
            data = json.loads(base64.b64decode(question["data"]).decode("utf-8").strip())

        logger.info("%r received a question.", self)

        question_uuid = get_nested_attribute(question, "attributes.question_uuid")
        forward_logs = bool(int(get_nested_attribute(question, "attributes.forward_logs")))
        return data, question_uuid, forward_logs
